#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>
#include "input.h"
#include "terminput.h"
#include "joystick.h"
#include "draw.h"
#include "tetris.h"
#include "socket.h"
#include "xterm.h"

const char input_chr9[9][4] = {
	"LFT", "RGT", "UP", "DWN", "(A)", "(B)", "BEL", "BS", "HT"
};

int inputdevs_fd[4] = {-1};
char inputdevs_player[4] = {0};

#ifdef JOYSTICK
char inputdevs_str[14];
static int pressedkeys;

int getinputdev(const char *key)
{
	char *p;
	if (key && (p = strchr(inputdevs_str, key[2])) && isdigit(*p))
		return *p-'0'+1;
	return 0;
}

const char *getinputdev_str(int i)
{
	char *p = inputdevs_str;
	while (i) {
		p = strchr(p, ' ');
		if (!p)
			return NULL;
		p++;
		i--;
	}
	return p;
}

static void init_inputdevs_str()
{
	strcpy(inputdevs_str, "stdin js0 js1");

	if (inputdevs_fd[1] == -1 || inputdevs_fd[2] == -1) {
		inputdevs_str[9] = '\0';
		if (inputdevs_fd[2] > -1)
			inputdevs_str[8] = '1';
		else if (inputdevs_fd[1] == -1)
			inputdevs_str[5] = '\0';
	}
}
#endif /* JOYSTICK */

void init_inputdevs()
{
	inputdevs_fd[0] = STDIN_FILENO;
#if NUM_INPUTDEVS > 1
	inputdevs_fd[1] = -1;
	inputdevs_fd[2] = -1;
	if (!socket_fd)
		inputdevs_fd[3] = -1;
#  ifdef JOYSTICK
	init_joysticks();
	init_inputdevs_str();
#  endif
#endif
}

#ifdef JOYSTICK
void close_inputdev(int i)
{
	close(inputdevs_fd[i]);
	inputdevs_fd[i] = -1;
	init_inputdevs_str();
}
#endif

static int setfds(fd_set *set)
{
#if NUM_INPUTDEVS > 1
	int n = 0;
	int i = NUM_INPUTDEVS-1;
	FD_ZERO(set);
	while (i >= 0) {
		if (inputdevs_fd[i] > -1) {
			FD_SET(inputdevs_fd[i], set);
			if (inputdevs_fd[i] >= n)
				n = inputdevs_fd[i]+1;
		}
		i--;
	}
	return n;
#else
	FD_ZERO(set);
	FD_SET(STDIN_FILENO, set);
	return STDIN_FILENO+1;
#endif
}

static void settimeout(struct timeval *tmv, int ms)
{
#ifdef JOYSTICK
	if (pressedkeys && ms > 5) {
		tmv->tv_sec = 0;
		tmv->tv_usec = 5000;
	} else {
		tmv->tv_sec = ms/1000;
		tmv->tv_usec = 1000*(ms%1000);
	}
#else
	tmv->tv_sec = ms/1000;
	tmv->tv_usec = 1000*(ms%1000);
#endif
}

static int inpselect(int n, fd_set *set, struct timeval *tmout)
{
	n = select(n, set, NULL, NULL, tmout);
	if (n == -1 && errno != EINTR) {
		perror("Error waiting for input");
		exit(1);
	}
	return n;
}

int gettm(int a)
{
	struct timeval tmv;
	int b;
	do {
		gettimeofday(&tmv, NULL);
		b = 1000*(tmv.tv_sec % 10) + tmv.tv_usec/1000;
		while (b < a)
			b += 10000;
	} while (a == b);
	return b;
}

int waitinput(int fd, unsigned us)
{
	fd_set set;
	struct timeval tmv;
	int ret;
	do {
		FD_ZERO(&set);
		FD_SET(fd, &set);
		tmv.tv_sec = 0;
		tmv.tv_usec = us;
		ret = inpselect(fd+1, &set, &tmv);
	} while (ret == -1);
	return ret;
}

#if defined(TWOPLAYER) && NUM_INPUTDEVS > 1
static int player_flag(int keypress, int i, int flags)
{
	if (keypress) {
		if (socket_fd > -1) {
			if (i==3)
				return PLAYER_2;
		} else if (inputdevs_player[i] && !(flags & SINGLE_PL))
			return inputdevs_player[i]==1 ? PLAYER_1 : PLAYER_2;
	}
	return 0;
}
#else
#define player_flag(k, i, flags) 0
#endif

static int getkeypress_dev(fd_set *set, int flags, int i)
{
	int fd = inputdevs_fd[i];
	int keypress = 0;
	if (fd > -1 && FD_ISSET(fd, set)) {
#ifdef SOCKET
		if (i==3) {
			keypress = sock_getkeypress(flags);
			goto out;
		}
#endif
#ifdef JOYSTICK
		if (i) {
			keypress = js_getpress(fd, i-1, flags);
			pressedkeys |= i;
			if (!js_pressed(i-1))
				pressedkeys &= ~i;
			goto out;
		}
#endif
		keypress = term_getkeypress(flags);
		if (flags & IN_GAME)
			switch (keypress & 63) {
			case MVDOWN:
			case HARDDROP:
				keypress |= SAFE_DROP;
			}
	}
#if NUM_INPUTDEVS > 1
out:	if (!keypress && i < NUM_INPUTDEVS-1)
		return getkeypress_dev(set, flags, i+1);
	keypress |= player_flag(keypress, i, flags);
#endif
	return keypress;
}

#ifdef JOYSTICK
static int getautorepeat(int flags)
{
	int keypress;
	int i = pressedkeys & 1;
	if (!i)
		i = 2;
	keypress = js_getautorepeat(i-1, flags);
	return keypress | player_flag(keypress, i, flags);
}

static void reset_autorepeat(int pl)
{
	int i;
	if (!pressedkeys)
		return;
	i = pressedkeys & 1;
	if (!i)
		i = 2;
	if (TWOPLAYER_MODE && pl==2)
		while (inputdevs_player[i] != 2) {
			if (i==2 || !(pressedkeys & 2))
				return;
			i = 2;
		}
	js_reset_autorepeat(i);
}
#else
#define reset_autorepeat(pl)
#endif

int getkeypress(int tm, int flags)
{
	fd_set set;
	struct timeval tmv;
	int t = gettm(0);
	int keypress;

#ifdef XTERMWIN
	if (tm > 50) {
		upd_termresize();
		if (in_xterm && game && game_running &&
		    !TWOPLAYER_MODE && !xterm_hasfocus())
			return STARTBTN;
	}
#endif
	if (socket_fd > -1)
		flags |= SINGLE_PL;
	while (1) {
		tm -= gettm(t)-t;
		if (tm <= 0)
			return 0;
		t = gettm(0);
		settimeout(&tmv, tm);
		if (inpselect(setfds(&set), &set, &tmv) > 0) {
			keypress = getkeypress_dev(&set, flags, 0);
			if (keypress)
				break;
		}
#ifdef JOYSTICK
		if (pressedkeys && (keypress = getautorepeat(flags)))
			break;
#endif
	}
	return keypress;
}

int getkeypress_block(int flags)
{
	int keypress;
	do {
#ifdef SOCKET
		if (invit || checkinvit())
			return ESC;
#endif
		keypress = getkeypress(1000, flags);
	} while (!keypress);
	return keypress;
}

#ifdef JOYSTICK
int test_autorep_tm(short *tm)
{
	int t = *tm;
	t = gettm(t)-t;
	if (t >= DAS_INITIAL_DELAY) {
		*tm = gettm(0)-t+DAS_DELAY+10000;
		return 1;
	}
	return 0;
}
#endif

static int key_is_valid(int c, int keypr)
{
	return c == (keypr & 0x7F) ||
		c > MVRIGHT &&
		(c > MVDOWN || keypr & IN_GAME) &&
		c != STARTBTN;
}

#ifdef JOYSTICK
static int setkeymapping_js(int dev, int keypr)
{
	int i = dev-1;
	int b = js_readbtn(inputdevs_fd[dev], i);
	if (!key_is_valid(b, keypr) || b < MVDOWN && keypr < MVDOWN &&
				       js_getbtnfor(i, keypr) < MVDOWN)
		return b;
	js_setmapping(i, b, keypr);
	return MVDOWN;
}
#endif

static int setkeymapping_term(int keypr)
{
	keytp *input;
	int c;
#ifndef CURSES
	char s[5] = "";
	if (c = term_readkey(s)) {
		c = c==1 ? (unsigned char) s[0] : ESC+1;
		input = s;
	}
#else
	if (c = term_readkey_int())
		input = &c;
#endif
	else
		return 0;
	if (!key_is_valid(c, keypr) || c == ESC || c == '\t' ||
	    c < MVDOWN && (keypr & (0x7F | IN_GAME)) < MVDOWN &&
	    c == term_getkeyfor_int(keypr, 1))
		return c;
	term_setmapping(input, keypr);
	return MVDOWN;
}

int setkeymapping(int dev, int keypr)
{
	fd_set set;
	struct timeval tmv;
	do {
#ifdef XTERMWIN
		upd_termresize();
#endif
		settimeout(&tmv, 5000);
	} while (inpselect(setfds(&set), &set, &tmv) <= 0);
#ifdef JOYSTICK
	if (inputdevs_fd[dev] == -1)
		return STARTBTN;
	if (!FD_ISSET(inputdevs_fd[dev], &set))
		return getkeypress_dev(&set, SINGLE_PL, 0) & 0xFF;
	if (dev)
		return setkeymapping_js(dev, keypr);
#endif
	return setkeymapping_term(keypr);
}

const char *getkeyfor_str(int dev, int keypr)
{
	static char s[7];
	int c;
#ifdef JOYSTICK
	if (dev) {
		c = js_getbtnfor(dev-1, keypr);
		if (c > '0') {
			sprintf(s, "%d", c-'0');
			return s;
		}
	} else
#endif
#ifndef CURSES
	switch (c = term_getkeyfor(keypr, s, 1)) {
	case 1: c = s[0];
	case 0: break;
	default:
		memmove(s+3, s+1, c);
		memcpy(s, "ESC", 3);
		return s;
	}
#else
	if ((c = term_getkeyfor_int(keypr, 1)) >= 0x100) {
		sprintf(s, "%o", c);
		return s;
	}
#endif
	if (!c)
		s[0] = '\0';
	else if (c == ' ')
		strcpy(s, "SPACE");
	else if (c == DEL)
		strcpy(s, "DEL");
	else if (c > ' ' || c < 0) {
		s[0] = c;
		s[1] = '\0';
	} else if (c < 10)
		return input_chr9[c-1];
	else
		sprintf(s, "%d", c);
	return s;
}

int processkey_ingame(int key, int flags)
{
	struct player *plr;
	int i = TWOPLAYER_MODE && key & PLAYER_2;
	int safe;
	switch (key & 0x7F) {
	case ESC:
	case BACKSPACE:
		return -2;
	case 'q':
		exit(0);
	case STARTBTN:
	case 'p':
		if (flags & NO_PAUSE || !game_running || TWOPLAYER_MODE)
			break;
		return pausegame();
	}
	if (flags & DISCARD_MOVES) {
		reset_autorepeat(1+i);
		return -1;
	}
	plr = game->player+i;
	safe = key & SAFE_DROP;
	key &= 0x7F;
	if (!(flags & DISCARD_DROPS)) {
		switch (key) {
		case HARDDROP:
			if (socket_fd > -1 && !i) {
				sock_sendpiece(plr);
				sock_sendbyte(HARDDROP);
				safe = 0;
			}
			return harddrop(plr, safe);
		case MVDOWN:
			if (!TWOPLAYER_MODE)
				return softdrop(softdrop_speed, safe);
			if (socket_fd > -1 && !i)
				sock_sendbyte(MVDOWN);
			return movedown(plr, 1);
		}
	}
	switch (key) {
	case MVLEFT:
		moveleft(plr);
		break;
	case MVRIGHT:
		moveright(plr);
		break;
	case MVUP:
		rotate(plr, plr->rotationsys & ROT_CLOCKWISE);
		break;
	case A_BTN:
		rotate(plr, 1);
		break;
	case B_BTN:
		rotate(plr, 0);
		break;
	default:
		return -1;
	}
	if (socket_fd > -1 && !i)
		sock_sendbyte(key);
	upd_screen(1+i);
	return 1;
}

void input_cleanup()
{
#if NUM_INPUTDEVS > 1
	int i = 1;
	if (inputdevs_fd[0] > -1) {
		while (i < NUM_INPUTDEVS) {
			if (inputdevs_fd[i] > -1)
				close(inputdevs_fd[i]);
			i++;
		}
		inputdevs_fd[0] = -1;
	}
	rminvitfile();
	rmsocket();
#endif
	free_escape_sequences();
}
