#include <string.h>
#include <curses.h>
#ifndef CURSES
#define CURSES
#endif
#include "term.h"
#include "draw.h"
#include "menu.h"
#include "tetris.h"

WINDOW *window;
WINDOW *wins[6] = {NULL};
static int margin_x;

static chtype acs_chars[17];

static void initpair(short pair, short f, short b)
{
	if (_TT_BLOCKS)
		b = -1;
	init_pair(pair, f, b);
}

static void init_color_pairs()
{
	int i;
	for (i = 1; i < 7; i++)
		initpair(i, i, i);
	if (_WHITE_BG)
		initpair(7, COLOR_BLACK, COLOR_BLACK);
	else
		initpair(7, COLOR_WHITE, COLOR_WHITE);
	init_pair(MAGENTA_FG, COLOR_MAGENTA, -1);
	initpair(WHITE_ON_BLUE, COLOR_WHITE, COLOR_BLUE);
	init_pair(BOARD_BG_COLOR, (_WHITE_BG) ? COLOR_CYAN : COLOR_BLUE, -1);
	init_pair(BOARD_FRAME_COLOR, COLOR_BLUE, -1);
	init_pair(RED_FG, COLOR_RED, -1);
	if (!_TT_BLOCKS) {
		init_pair(15, COLOR_GREEN, -1);
		init_pair(16, COLOR_YELLOW, -1);
		init_pair(17, COLOR_CYAN, -1);
	}
}

void term_setup()
{
	initscr();
	cbreak();
	noecho();
	keypad(stdscr, TRUE);
	curs_set(0);
	if (!_MONOCHROME && start_color() != ERR && COLOR_PAIRS >= 16) {
#ifdef NCURSES_VERSION
		use_default_colors();
#endif
		init_color_pairs();
	} else
		term_flags |= MONOCHROME;
	window = stdscr;
	refresh();
	acs_chars[0] = ACS_BOARD;	/* h */
	acs_chars[2] = ACS_LRCORNER;	/* j */
	acs_chars[3] = ACS_URCORNER;	/* k */
	acs_chars[4] = ACS_ULCORNER;	/* l */
	acs_chars[5] = ACS_LLCORNER;	/* m */
	acs_chars[6] = ACS_PLUS;	/* n */
	acs_chars[9] = ACS_HLINE;	/* q */
	acs_chars[12] = ACS_LTEE;	/* t */
	acs_chars[13] = ACS_RTEE;	/* u */
	acs_chars[14] = ACS_BTEE;	/* v */
	acs_chars[15] = ACS_TTEE;	/* w */
	acs_chars[16] = ACS_VLINE;	/* x */
}

static void delwins()
{
	int i;
	for (i = 0; i < 6; i++)
		if (wins[i]) {
			delwin(wins[i]);
			wins[i] = NULL;
		}
}

void restoreterm()
{
	int x, y;
	delwins();
	if (!isendwin()) {
		getmaxyx(stdscr, y, x);
		move(y-1, 0);
		clrtobot();
		refresh();
		endwin();
	}
}

void setcurs(int x, int y)
{
	if (window == stdscr)
		x += margin_x;
	else if (x < 0) {
		margin_x = x;
		x = 0;
	} else
		margin_x = 0;
	wmove(window, y, x);
}

static int winindex(int win)
{
	switch (win) {
	case WIN_TETROM_STATS:
		return 2;
	case WIN_TOP_SCORES:
		return 4;
	}
	return win;
}

void setwcurs(int win, int x, int y)
{
	window = wins[winindex(win)];
	if (!window) {
		window = stdscr;
		margin_x = getmargin_x();
	}
	setcurs(x, y);
}

void movefwd(int n)
{
	int x, y;
	getyx(window, y, x);
	x += n;
	if (margin_x < 0)
		setcurs(x+margin_x, y);
	else
		wmove(window, y, x);
}

void newln(int x)
{
	int d, y;
	getyx(window, y, d);
	setcurs(x, y+1);
}

void setcurs_end()
{
	move(term_height-1, 0);
	refresh();
	window = stdscr;
}

int is_outside_screen(int x)
{
	int max, y;
	if (window == stdscr)
		x += margin_x;
	getmaxyx(window, y, max);
	return x >= max;
}

#ifndef NO_MENU
void get_xy(int *x, int *y)
{
	getyx(window, *y, *x);
}
#endif

void refreshwin(int i)
{
	WINDOW *win;
	if (i == -1)
		win = window;
	else
		win = wins[winindex(i)];
	if ((i==1 || i==2) && !game_running)
		touchwin(win);
	if (win)
		wrefresh(win);
	else {
		attrset(A_NORMAL);
		refresh();
	}
}

void clearwin(int win)
{
	setwcurs(win, 0, 0);
	werase(window);
}

void cleartoeol()
{
	wclrtoeol(window);
}

static void setfgcolor(int clr)
{
	int pairs[6] = {RED_FG, 15, 16, BOARD_FRAME_COLOR, MAGENTA_FG, 17};
	int attrs = A_BOLD;
	clr &= 7;
	if (_TT_BLOCKS)
		attrs |= COLOR_PAIR(clr);
	else if (clr < 0x17)
		attrs |= COLOR_PAIR(pairs[clr-1]);
	wattrset(window, attrs);
}

static void set_panel_label_color()
{
	short f = COLOR_YELLOW;
	short b;
	if (TWOPLAYER_MODE)
		b = COLOR_BLUE;
	else {
		b = (player1.level % 6)+1;
		if (b == COLOR_CYAN)
			f = COLOR_WHITE;
	}
	initpair(PANEL_LABEL_COLOR, f, b);
}

void setcolorpair(int pair)
{
	int attrs;
	if (_MONOCHROME) {
		if (pair == MAGENTA_FG)
			setattr_bold();
		return;
	}
	if (pair == YELLOW_ON_BLUE) {
		initpair(YELLOW_ON_GREEN, COLOR_YELLOW, COLOR_BLUE);
		pair = YELLOW_ON_GREEN;
	} else if (pair == YELLOW_ON_GREEN)
		initpair(YELLOW_ON_GREEN, COLOR_YELLOW, COLOR_GREEN);
	else if (pair & 16) {
		setfgcolor(pair);
		return;
	}
	attrs = COLOR_PAIR(pair);
	switch (pair) {
	case BOARD_BG_COLOR:
	case BOARD_FRAME_COLOR:
		break;
	case PANEL_LABEL_COLOR:
		set_panel_label_color();
		attrs |= A_BOLD;
		break;
	case RED_FG:
		if (_WHITE_BG)
			break;
	default:
		attrs |= A_BOLD;
	}
	wattrset(window, attrs);
}

void setattr_normal()
{
	wattrset(window, A_NORMAL);
}

void setattr_standout()
{
	wstandout(window);
}

void setattr_bold()
{
	if ((term_flags & TT_MONO) != TT_MONO)
		wattron(window, A_BOLD);
}

void setattr_underline()
{
	wattron(window, A_UNDERLINE);
}

static int putch_acs1(int ch)
{
	chtype c;
	switch (ch) {
	case TEXTURE1:
		c = ACS_CKBOARD;
		break;
	case BULLET:
		c = ACS_BULLET;
		break;
	case UPARROW:
		c = ACS_UARROW;
		break;
	default:
		return 0;
	}
	waddch(window, c);
	return 1;
}

static int putch_acs(int ch)
{
	int x, y;
	int i;
	if (putch_acs1(ch))
		return 1;
	if (ch >= TEXTURE2 && ch <= VLINE) {
		i = ch-0x100-'h';
		if (acs_chars[i]) {
			getyx(window, y, x);
			waddch(window, acs_chars[i]);
			if (is_outside_screen(x+1))
				wmove(window, y, x);
			return 1;
		}
	}
	return 0;
}

void putch(int ch)
{
	if (!(ch & 0x100)) 
		waddch(window, ch);
	else if (_ASCII || ch == (' '|0x100) || ch == TEXTURE2 &&
				term_flags & (XTERM | TT_BLOCKS))
		putch_ascii(ch);
	else if (!putch_acs(ch))
		waddch(window, ch & ~0x100);
}

int printstr(const char *str)
{
	int x, y, w;
	int n = strlen(str);
	getyx(window, y, x);
	getmaxyx(window, y, w);
	if (x+n >= w)
		n = w-x-1;
	waddnstr(window, str, n);
	return n;
}

void printint(const char *fmt, int d)
{
	wprintw(window, fmt, d);
}

int default_bgdot()
{
	return _XTERM ? BULLET : '.';
}

#ifndef NO_MENU
void term_entermenu()
{
	int x = getmargin_x();
	delwins();
	refresh();
	window = newwin(term_height-1, term_width-x, 1, x);
	wins[0] = window;
	draw_tetris_logo(0, 0);
	print_vitetris_ver(19, 4);
}

static void printmenuhelp(int y)
{
	if (term_height < y+5 || term_width < 48)
		return;
	setcurs(0, y);
	printstr("Use the arrow keys or TAB to navigate,");
	newln(0);
	printstr("ENTER to proceed, BACKSPACE or ESC to");
	newln(0);
	printstr("go back.  Exit at once with Q.");
}

static void draw_options_box()
{
	drawbox(1, 7, 31, 10+!(term_flags & CYGWIN), "Options");
}

static int op_handler(int k, int *pos)
{
	const struct termopt opts[2] = {
		{ WHITE_BG, "black white", "bg" },
		{ ASCII, "VT100 ASCII", "drawing" }
	};
	int i = *pos-2;
	if (i < 0) {
		if (k) {
			setcurs(0, 16);
			wclrtobot(window);
			inputsetup_box(k-1, 1, 7);
		}
		draw_options_box();
		return 1;
	}
	if (i < 1+!(term_flags & CYGWIN))
		i = term_optionhandler(k, opts+i);
	else {
		if (!k) {
			getyx(window, i, k);
			setcurs(2, i+4);
			putnchars(HLINE, 28);
			setcurs(k, i);
			k = 0;
		}
		i = term_blockstyle(k);
	}
	if (i == 3) {
		draw_tetris_logo(0, 0);
		if (!_MONOCHROME) {
			init_color_pairs();
			if (*pos == 2)
				clearok(window, TRUE);
		}
		draw_options_box();
	}
	return i;
}

static void term_optionsmenu()
{
	const char *items[2] = {"Term BG", "Line Drawing"};
	int n = 1+!(term_flags & CYGWIN);
	draw_options_box();
	optionsmenu(items, n, op_handler, 3, 9);
	setcurs(0, 7);
	wclrtobot(window);
}

static void hiscorelist()
{
	int y = 7;
	if (term_height < 22)
		y--;
	show_hiscorelist(1, y);
	clearbox(0, y, 0, 12);
}

#ifdef TWOPLAYER
int term_startupmenu(int i)
{
	while (1) {
		drawbox(5, 7, 22, 10, NULL);
		printmenuhelp(19);
		i = startupmenu(i-1, 8, 9);
# ifndef SOCKET
		if (!i)
			break;
# else
		if (i==2 && !select_2p_tty(24, 10))
			continue;
		if (!i) {
			setcurs(5, 14);
			wclrtobot(window);
			if ((i = menu_checkinvit(1, 7)) == -1) {
				refresh();
				setwcurs(0, 5, 7);
				i = 1;
				continue;
			}
			if (!i)
				break;
		}
		move(term_height-1, 0);
		clrtoeol();
		refresh();
# endif
		setcurs(5, 7);
		wclrtobot(window);
		if (i==5)
			term_optionsmenu();
		else if (i==6)
			hiscorelist();
		else
			break;
	}
	return i;
}
#endif /* TWOPLAYER */

int term_gamemenu()
{
	int i = 0;
#ifdef TWOPLAYER
	int y = 7;
	while (1) {
		if (!TWOPLAYER_MODE) {
			if (term_height >= 23)
				show_hiscorelist5(5, GAMEMENU_LENGTH+9, 0);
		} else if (term_height < 23) {
			setcurs(1, 4);
			cleartoeol();
			y = 5;
		}
		i = gamemenu(i, 1, y);
		if (i != 4)
			break;
		wclrtobot(window);
		inputsetup_box(0, 1, 7);
		i = 3;
	}
	if (!i) {
		setcurs(1, 4);
		wclrtobot(window);
		print_vitetris_ver(19, 4);
	}
#else
	while (1) {
		printmenuhelp(19);
		i = gamemenu(i, 1, 7);
		if (!i)
			break;
		clearbox(0, 7, 0, GAMEMENU_LENGTH);
		if (i != 4 && i < GAMEMENU_LENGTH-1)
			break;
		wclrtobot(window);
		switch (i) {
		case 4:
			inputsetup_box(0, 1, 7);
			break;
		case GAMEMENU_LENGTH-1:
			term_optionsmenu();
			break;
		default:
			hiscorelist();
		}
		i--;
	}
#endif
	return i;
}
#endif /* !NO_MENU */

static void createwin(int i, int w, int h)
{
	int x, y;
	getwin_xy(i, &x, &y);
	i = winindex(i);
	wins[i] = newwin(h, w, y, x+margin_x);
}

void term_entergame()
{
	delwins();
	margin_x = getmargin_x();
	createwin(1, 20, 20);
	if (!TWOPLAYER_MODE) {
		createwin(WIN_NEXT, 8, 2);
		createwin(WIN_PANEL, 10, 20);
		if (term_width >= 45) {
			createwin(WIN_TETROM_STATS, 9, 9);
			if (term_width >= 47)
				createwin(WIN_TOP_SCORES, 11, 7);
		}
	} else {
		createwin(2, 20, 20);
		if (_HEIGHT_24L || term_width >= 76) {
			createwin(WIN_NEXT, 8, 2);
			createwin(WIN_NEXT+1, 8, 2);
		}
		createwin(WIN_PANEL, 12, 20);
	}
	clear();
}
